from .utils import pairwise_dist as pdist
import numpy as np

def measure(orig, emb, sigma=0.1, distance_matrices=None):
	"""
	Compute the distance to measure of the embedding
	INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
		tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
	OUTPUT:
		dict: distance to measure (dtm)
	"""

	if distance_matrices is None:
		orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
		emb_distance_matrix = pdist.pairwise_distance_matrix(emb)
	else:
		orig_distance_matrix, emb_distance_matrix = distance_matrices

	density_orig = pdist.distance_matrix_to_density(orig_distance_matrix, sigma)
	density_emb = pdist.distance_matrix_to_density(emb_distance_matrix, sigma)

	dtm = np.abs(density_orig - density_emb).sum()

	return {
		"distance_to_measure": dtm
	}


import numpy as np

def measure(emb, label):
  """
	Compute distance consistency of the embedding
	INPUT:
		ndarray: emb: embedded data
		ndarray: label: label of the original data
	OUTPUT:
		dict: distance consistency (dc)
  """
  
	## compute centroids
  point_num = emb.shape[0]
  label_num = np.unique(label).shape[0]

  centroids = np.zeros((label_num, emb.shape[1]))
  for i in range(label_num):
    centroids[i] = np.mean(emb[label == i], axis=0)
    
	## compute distance consistency
  consistent_num = 0
  for idx in range(point_num):
    current_label = -1
    current_dist = 1e10
    for c_idx in range(len(centroids)):
      dist = np.linalg.norm(emb[idx] - centroids[c_idx])
      if dist < current_dist:
        current_dist = dist
        current_label = c_idx
    if current_label == label[idx]:
      consistent_num += 1
	
  dsc = consistent_num / point_num
  return {
    "distance_consistency": dsc
	}


from sklearn.neighbors import kneighbors_graph
import numpy as np
from .utils import knn



def measure(orig, emb, k=20, snn_info=None, knn_info=None):
	"""
	Compute neighbor dissimilarity (ND) of the embedding
	INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
		int: k: number of nearest neighbors to consider
		tuple: knn_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
	OUTPUT:
		dict: neighbor dissimilarity (ND)
	"""
	
	if snn_info is None:
		if knn_info is None:
			orig_SNN_graph = knn.snn(orig, k, directed=False)
			emb_SNN_graph = knn.snn(emb, k, directed=False)
		else:
			orig_SNN_graph = knn.snn(orig, k, knn_indices=knn_info[0], directed=False)
			emb_SNN_graph = knn.snn(emb, k, knn_indices=knn_info[1], directed=False)
	
	else:
		orig_SNN_graph, emb_SNN_graph = snn_info

	D = (orig_SNN_graph - emb_SNN_graph) / k

	D_plus = D[D > 0]
	D_minus = D[D < 0]

	dissim_plus = np.sqrt(np.sum(D_plus**2))
	dissim_minus = np.sqrt(np.sum(D_minus**2))

	nd = max(dissim_plus, dissim_minus)

	return {
			"neighbor_dissimilarity": nd
	}


__all__ = [
    'class_aware_trustworthiness_continuity', 'spearman_rho', 'neighbor_dissimilarity', 
		'distance_to_measure', 'local_continuity_meta_criteria', 'internal_validation_measure', 
		'pearson_r', 'distance_consistency', 'kl_divergence', 'neighborhood_hit', 
		'trustworthiness_continuity', 'clustering_and_external_validation_measure', 'mean_relative_rank_error', "steadiness_cohesiveness",
		'topographic_product', 'procrustes', 'stress'
]

## add to __all__ if the function is added


from sklearn.cluster import KMeans, DBSCAN
from sklearn.metrics import adjusted_rand_score, adjusted_mutual_info_score, normalized_mutual_info_score, v_measure_score


def measure(emb, label, measure="arand",  clustering="kmeans", clustering_args=None):
	"""
	Evaluate DR embedding using clustering and external validation measure
	INPUT:
		ndarray: emb: embedded data
		str: clustering: clustering algorithm to use (Optional)
			Currently supports "K-Means", "DBSCAN"
		dict: clustering_args: arguments for clustering algorithm (Optional)
		str: measure: external validation measure to compute (Optional)
			Currently supports "arand", "ami", "nmi", "vmeasure"
	OUTPUT:
		dict: clustering and external validation measure value
	"""

	if clustering_args is None:
		clustering_args = {}

	if clustering == "kmeans":

		clustering_result = KMeans(**clustering_args).fit(emb)
	elif clustering == "dbscan":
		clustering_result = DBSCAN(**clustering_args).fit(emb)
	else:
		raise ValueError("Invalid clustering algorithm")

	if measure == "arand":
		score = adjusted_rand_score(label, clustering_result.labels_)
	elif measure == "ami":
		score = adjusted_mutual_info_score(label, clustering_result.labels_)
	elif measure == "nmi":
		score = normalized_mutual_info_score(label, clustering_result.labels_)
	elif measure == "vmeasure":
		score = v_measure_score(label, clustering_result.labels_)
	
	return {
		f"{clustering}_{measure}": score
	}



from scipy.spatial.distance import cdist
import numpy as np

def pairwise_distance_matrix(point, distance_function="euclidean"):
	"""
	Compute the pairwise distance matrix of the point list
	You can use any distance function from scipy.spatial.distance.cdist or specify a callable function
	INPUT:
		ndarray: point: list of points
		str or callable: distance_function: distance function to use
	OUTPUT:
		ndarry: pairwise distance matrix 
	"""
	if callable(distance_function):
		distance_matrix = cdist(point, point, distance_function)
	elif distance_function == "snn":
		## TODO
		pass
	else:
		distance_matrix = cdist(point, point, distance_function)
	return distance_matrix

def distance_matrix_to_density(distance_matrix, sigma):
	"""
	Compute the density of each point based on the pairwise distance matrix
	INPUT:
		ndarray: distance_matrix: pairwise distance matrix
		float: sigma: sigma parameter for the Gaussian kernel
	OUTPUT:
		ndarry: density
	"""

	normalized_distance_matrix = distance_matrix / np.max(distance_matrix)
	density = np.sum(np.exp(- (normalized_distance_matrix ** 2) / sigma), axis=-1)
	density = density / np.sum(density)
	return density


import numpy as np 
from .utils import knn

def measure(orig, emb, k=20, knn_ranking_info=None, return_local=False):
	"""
	Compute the trustworthiness and continuity of the embedding
	INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
		int: k: number of nearest neighbors to consider
		tuple: knn_ranking_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
	OUTPUT:
		dict: trustworthiness and continuity
	"""
	if knn_ranking_info is None:
		orig_knn_indices, orig_ranking = knn.knn_with_ranking(orig, k)
		emb_knn_indices,  emb_ranking  = knn.knn_with_ranking(emb, k)
	else:
		orig_knn_indices, orig_ranking, emb_knn_indices, emb_ranking = knn_ranking_info

	if return_local:
		trust, local_trust = tnc_computation(orig_knn_indices, orig_ranking, emb_knn_indices, k, return_local)
		cont , local_cont  = tnc_computation(emb_knn_indices,  emb_ranking, orig_knn_indices, k, return_local)
		return ({
			"trustworthiness": trust,
			"continuity": cont
		}, {
			"local_trustworthiness": local_trust,
			"local_continuity": local_cont
		})
	else:
		trust = tnc_computation(orig_knn_indices, orig_ranking, emb_knn_indices, k, return_local)
		cont  = tnc_computation(emb_knn_indices,  emb_ranking, orig_knn_indices, k, return_local)
		return {
			"trustworthiness": trust,
			"continuity": cont
		}

def tnc_computation(base_knn_indices, base_ranking, target_knn_indices, k, return_local=False):
	"""
	Core computation of trustworthiness and continuity
	"""
	local_distortion_list = []
	points_num = base_knn_indices.shape[0]

	for i in range(points_num):
		missings = np.setdiff1d(target_knn_indices[i], base_knn_indices[i])
		local_distortion = 0.0 
		for missing in missings:
			local_distortion += base_ranking[i, missing] - k
		local_distortion_list.append(local_distortion)
	local_distortion_list = np.array(local_distortion_list)
	local_distortion_list = 1 - local_distortion_list * (2 / (k * (2 * points_num - 3 * k - 1)))

	average_distortion = np.mean(local_distortion_list)

	if return_local:
		return average_distortion, local_distortion_list
	else:
		return average_distortion
	

	

	


from .utils import pairwise_dist as pdist
import numpy as np

def measure(orig, emb, distance_matrices=None):
	"""
	Compute stress of the embedding
	INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
        tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
	OUTPUT:
		dict: stress
	"""
	if distance_matrices is None:
		orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
		emb_distance_matrix = pdist.pairwise_distance_matrix(emb)

	else:
		orig_distance_matrix, emb_distance_matrix = distance_matrices

	diff_squared_sum = np.square(orig_distance_matrix - emb_distance_matrix).sum()
	orig_squared_sum = np.square(orig_distance_matrix).sum()

	stress = np.sqrt(diff_squared_sum / orig_squared_sum)
	
	return {
		"stress": stress
	}


from .utils import pairwise_dist as pdist
import numpy as np
from .utils import knn

def measure(orig, emb, k=20, distance_matrices=None, knn_info=None):
	"""
	Compute topographic product
	INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
	      int: k: number of nearest neighbors to consider
	OUTPUT:
		topographic product result
	"""
	N = len(emb)
	sum_of_log_p3 = 0
	
	if distance_matrices is None:
		
		orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
		emb_distance_matrix  = pdist.pairwise_distance_matrix(emb)
	else:
		orig_distance_matrix, emb_distance_matrix = distance_matrices

	# k nearest neighbors in original space and embedded space each
	if knn_info is None:
		orig_knn_indices = knn.knn(orig, k)
		emb_knn_indices  = knn.knn(emb, k)
	else:
		orig_knn_indices, emb_knn_indices = knn_info
	
	for j in range(N):
		for k in range(N-1):
			q1_product, q2_product = 1, 1
			for l in range(k):
				distance_origin_to_emb_knn      = orig_distance_matrix[j][emb_knn_indices[j][l]]
				distance_origin_to_origin_knn   = orig_distance_matrix[j][orig_knn_indices[j][l]]
				q1 = distance_origin_to_emb_knn / distance_origin_to_origin_knn
				q1_product *= q1

				distance_emb_to_emb_knn        = emb_distance_matrix[j][emb_knn_indices[j][l]]
				distance_emb_to_origin_knn     = emb_distance_matrix[j][orig_knn_indices[j][l]]
				q2 = distance_emb_to_emb_knn   / distance_emb_to_origin_knn
				q2_product *= q2

			p3 = pow(q1_product * q2_product, 1 / (2 * k))
			sum_of_log_p3 += np.log(p3)

	return sum_of_log_p3 / (N * (N - 1))
            


from snc.snc import SNC
import numpy as np
from .utils import knn



def measure(orig, emb, iteration=150, walk_num_ratio=0.3, alpha=0.1, k=50, clustering_strategy="dbscan", knn_info=None, return_local=False):
	"""
	Compute the Steadiness and Cohesiveness of the embedding
	"""

	if knn_info is None:
		orig_knn_indices = knn.knn(orig, k)
		emb_knn_indices = knn.knn(emb, k)
	else:
		orig_knn_indices, emb_knn_indices = knn_info
	
	orig_snn_graph = knn.snn(orig, k, knn_indices=orig_knn_indices, directed=True)
	emb_snn_graph = knn.snn(emb, k, knn_indices=emb_knn_indices, directed=True)

	snn_knn_matrix = {
		"raw_knn": orig_knn_indices,
		"raw_snn": orig_snn_graph,
		"emb_knn": emb_knn_indices,
		"emb_snn": emb_snn_graph
	}


	snc_obj = SNC(
		orig, emb, 
		iteration=iteration, 
		walk_num_ratio=walk_num_ratio, 
		dist_strategy="inject_snn", 
		dist_parameter={ "alpha": alpha }, 
		dist_function=None, 
		cluster_strategy=clustering_strategy, 
		snn_knn_matrix=snn_knn_matrix
	)

	snc_obj.fit(record_vis_info=return_local)

	steadiness = snc_obj.steadiness()
	cohesiveness = snc_obj.cohesiveness()

	if return_local:
		_, _, _, points_info = snc_obj.vis_info()

		stead_local = [1 - point_info["false_val"] for point_info in points_info]
		cohev_local = [1 - point_info["missing_val"] for point_info in points_info]

	if return_local:
		return { ## TODO
			"steadiness": steadiness,
			"cohesiveness": cohesiveness
		}, {
			"local_steadiness": stead_local,
			"local_cohesiveness": cohev_local
		}
	else:
		return {
			"steadiness": steadiness,
			"cohesiveness": cohesiveness
		}



from .pairwise_dist import pairwise_distance_matrix
import numpy as np
import faiss
from sklearn.neighbors import KDTree

def knn_with_ranking(points, k, distance_matrix=None):
  """
  Compute the k-nearest neighbors of the points along with the 
  rankings of other points based on the distance to each point.
  If the distance matrix is not provided, it is computed in O(n^2) time.
  INPUT:
		ndarray: points: list of points
		int: k: number of nearest neighbors to compute
		ndarray: distance_matrix: pairwise distance matrix (Optional)
  OUTPUT:
		ndarray: knn_indices: k-nearest neighbors of each point 
		ndarray: ranking: ranking of other points based on the distance to each point
  """
  
  if distance_matrix is None:
    distance_matrix = pairwise_distance_matrix(points, "euclidean")

  knn_indices = np.empty((points.shape[0], k), dtype=np.int32)
  ranking = np.empty((points.shape[0], points.shape[0]), dtype=np.int32)
  
  for i in range(points.shape[0]):
    distance_to_i = distance_matrix[i]
    sorted_indices = np.argsort(distance_to_i)
    knn_indices[i] = sorted_indices[1:k+1]
    ranking[i] = np.argsort(sorted_indices)
  
  return knn_indices, ranking
  

def knn(points, k, distance_function="euclidean"):
  """
  Compute the k-nearest neighbors of the points
  If the distance function is euclidean, the computation relies on faiss-cpu.
  Otherwise, the computation is done based on scikit-learn KD Tree algorithm
  You can use any distance function supported by scikit-learn KD Tree or specify a callable function
  INPUT:
		ndarray: points: list of points
		int: k: number of nearest neighbors to compute
		str or callable: distance_function: distance function to use
  OUTPUT:
		ndarray: knn_indices: k-nearest neighbors of each point 
	"""
	
	## make c-contiguous
  points = np.ascontiguousarray(points, dtype=np.float32)

  if distance_function == "euclidean":
    index = faiss.IndexFlatL2(points.shape[1])
    index.add(points)
    knn_indices = index.search(points, k+1)[1][:, 1:]
  else:
    tree = KDTree(points, metric=distance_function)
    knn_indices = tree.query(points, k=k+1, return_distance=False)[:, 1:]
	
  return knn_indices

def snn(points, k, distance_function="euclidean", directed=True, knn_indices=None):
  """
	Compute the shared nearest neighbors (SNN) graph of the points
	INPUT:
		ndarray: points: list of points
		int: k: number of nearest neighbors to consider
		str or callable: distance_function: distance function to use
    bool: directed: whether the k-nearest neighbors graph using is directed or not
		tuple: knn_info: precomputed k-nearest neighbors and rankings of the points (Optional)
	OUTPUT:
		ndarray: snn_graph: shared nearest neighbors (SNN) graph of the points
	"""
  if knn_indices is None:
    knn_indices = knn(points, k, distance_function)

  knn_graph = np.zeros((knn_indices.shape[0], knn_indices.shape[0]))
  for i in range(k):
    knn_graph[np.arange(knn_indices.shape[0]), knn_indices[:, i]] = k-i

  if directed:
    snn_graph = knn_graph @ knn_graph.T
  else:
    knn_graph = ((knn_graph + knn_graph.T) > 0).astype(float)
    snn_graph = knn_graph @ knn_graph

  np.fill_diagonal(snn_graph, 0)

  return snn_graph


from .measures import *
from .measures.utils import knn
from .measures.utils import pairwise_dist as pdist

class ZADU:

	ABBREVIATIONS = {
		"tnc": "trustworthiness_continuity",
		"mrre": "mean_relative_rank_error",
		"lcmc": "local_continuity_meta_criteria",
		"nh": "neighborhood_hit",
		"ca_tnc": "class_aware_trustworthiness_continuity",
		"nd": "neighbor_dissimilarity",
		"dtm": "distance_to_measure",
		"lcmc": "local_continuity_meta_criteria",
		"kl_div": "kl_divergence",
		"dtm": "distance_to_measure",
		"dsc": "distance_consistency",
		"pr": "pearson_r",
		"srho":	"spearman_rho",
		"ivm": "internal_validation_measure",
		"c_evm": "clustering_and_external_validation_measure",
		"snc": "steadiness_cohesiveness",
		"topo": "topographic_product",
		"proc": "procrustes",
		"stress": "stress"
	}

	DEFAULT_K = 20
	
	def __init__(self, spec_list, orig, return_local=False, verbose=False):
		self.spec_list    = spec_list
		self.return_local = return_local
		self.verbose      = verbose

		self.orig  = orig
		self.emb   = None
		self.label = None

		## FLAGS for scheduling
		self.knn_flag   						= False
		self.knn_flag_k 						= -1
		self.knn_ranking_flag 			= False
		self.knn_ranking_flag_k 		= -1
		self.distance_matrices_flag = False

		## variables for holding precomputed results (prerequisite for some measures)
		self.orig_distance_matrix = None
		self.emb_distance_matrix  = None
		self.orig_knn_ranking     = None
		self.emb_knn_ranking      = None
		self.orig_knn_indices		  = None
		self.emb_knn_indices		  = None

		self.__sanity_check_measures_spec()
		self.__interpret_measures_spec()

		if self.distance_matrices_flag:
			self.orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
		if self.knn_ranking_flag:
			self.orig_knn_indices, self.orig_knn_ranking = knn.knn_with_ranking(orig, self.knn_ranking_flag_k, distance_matrix=self.orig_distance_matrix)
		elif self.knn_flag and self.knn_flag_k > self.knn_ranking_flag_k:
			self.orig_knn_indices = knn.knn(orig, self.knn_flag_k, distance_function="euclidean")
	

	def measure(self, emb, label=None):
		"""
		Run the functions specified in spec_list
		INPUT:
			emb:  embedded data
		OUTPUT:
			list: list of results
		"""

		self.emb  = emb
		self.label = label

		## compute the distance matrices
		if self.distance_matrices_flag:
			self.emb_distance_matrix  = pdist.pairwise_distance_matrix(emb)
		if self.knn_ranking_flag:
			self.emb_knn_indices,  self.emb_knn_ranking  = knn.knn_with_ranking(emb,  self.knn_ranking_flag_k, distance_matrix=self.emb_distance_matrix)
		elif self.knn_flag and self.knn_flag_k > self.knn_ranking_flag_k:
			self.emb_knn_indices  = knn.knn(emb,  self.knn_flag_k, distance_function="euclidean")
		
		## compute the measures
		score_results = []
		local_results = []
		for spec in self.spec_list:
			measure_name = spec["id"]
			given_params = spec["params"] if "params" in spec else {}
			real_params  = self.__get_real_params(measure_name)

			## construct the execution parameters to be injected in the function
			exec_params = {}
			for param in given_params.keys():
				exec_params[param] = given_params[param]

			for param in real_params:
				if "orig" == param:
					exec_params["orig"] = self.orig
				elif "emb" == param:
					exec_params["emb"] = self.emb
				elif "label" == param:
					if label is None:
						raise Exception("Label is required for measure {}".format(measure_name))
					exec_params["label"] = label
				elif "distance_matrices" == param:
					exec_params["distance_matrices"] = (self.orig_distance_matrix, self.emb_distance_matrix)
				elif "knn_ranking_info" == param:
					k_val = exec_params["k"] if "k" in exec_params else self.DEFAULT_K
					exec_params["knn_ranking_info"] = (
						self.orig_knn_indices[:, :k_val], 
						self.orig_knn_ranking, 
						self.emb_knn_indices[:, :k_val], 
						self.emb_knn_ranking
					)
				elif "knn_indices" == param:
					k_val = exec_params["k"] if "k" in exec_params else self.DEFAULT_K
					exec_params["knn_indices"] = (
						self.orig_knn_indices[:, :k_val], 
						self.emb_knn_indices[:, :k_val]
					)
				elif "return_local" == param:
					exec_params["return_local"] = self.return_local
			
			## execute the function
			if self.return_local and "return_local" in exec_params:
				score, local = globals()[measure_name].measure(**exec_params)
				score_results.append(score)
				local_results.append(local)
			elif self.return_local and "return_local" not in exec_params:
				score = globals()[measure_name].measure(**exec_params)
				score_results.append(score)
				local_results.append(None)
			else:
				score = globals()[measure_name].measure(**exec_params)
				score_results.append(score)

		if self.return_local:
			return score_results, local_results
		else:
			return score_results
		



	def __sanity_check_measures_spec(self):
		"""
		Perform sanity check on the measures specification list.
		"""
		## check whehter there exists invalid measure name
		for spec in self.spec_list:
			if spec["id"] not in self.ABBREVIATIONS.values():
				if spec["id"] in self.ABBREVIATIONS:
					spec["id"] = self.ABBREVIATIONS[spec["id"]]
				else:
					raise Exception("Invalid measure name: {}".format(spec["id"]))

		## check whether the parameters are valid
		for spec in self.spec_list:
			measure_name = spec["id"]
			given_params = spec["params"] if "params" in spec else {}
			real_params  = self.__get_real_params(measure_name)

			## check whether the given parameters are valid
			for param in given_params:
				if param not in real_params:
					raise Exception(f"Invalid parameter {param} for measure {measure_name}")
	


	def __interpret_measures_spec(self):
		"""
		Interpret the measures spec and specify the preprequisites (knn, distance matrices)
		"""
		for spec in self.spec_list:
			measure_name = spec["id"]
			given_params = spec["params"] if "params" in spec else {}
			real_params  = self.__get_real_params(measure_name)

			if "knn_ranking_info" in real_params:
				self.knn_ranking_flag = True
				if "k" in given_params:
					self.knn_ranking_flag_k = max(self.knn_ranking_flag_k, given_params["k"])
				else:
					self.knn_ranking_flag = max(self.knn_ranking_flag_k, self.DEFAULT_K)
			if "knn_info" in real_params:
				self.knn_flag = True
				if "k" in given_params:
					self.knn_flag_k = max(self.knn_flag_k, given_params["k"])
				else:
					self.knn_flag = max(self.knn_flag_k, self.DEFAULT_K)
			if "distance_matrices" in real_params:
				self.distance_matrices_flag = True



	
	def __get_real_params(self, measure_name):
		"""
		Get the real parameters of a measure.
		"""
		measure_func = globals()[measure_name].measure
		return measure_func.__code__.co_varnames[:measure_func.__code__.co_argcount]


import numpy as np


def cielab_to_rgb_hex(L, a, b):
	# Convert CIELAB to XYZ
	def lab_to_xyz(L, a, b):
		y = (L + 16) / 116
		x = a / 500 + y
		z = y - b / 200

		x = x ** 3 if x ** 3 > 0.008856 else (x - 16 / 116) / 7.787
		y = y ** 3 if y ** 3 > 0.008856 else (y - 16 / 116) / 7.787
		z = z ** 3 if z ** 3 > 0.008856 else (z - 16 / 116) / 7.787

		x = x * 95.047
		y = y * 100
		z = z * 108.883

		return x, y, z

	# Convert XYZ to RGB
	def xyz_to_rgb(x, y, z):
		x /= 100
		y /= 100
		z /= 100

		r = x * 3.2406 + y * -1.5372 + z * -0.4986
		g = x * -0.9689 + y * 1.8758 + z * 0.0415
		b = x * 0.0557 + y * -0.2040 + z * 1.0570

		def convert(color):
			color = np.clip(color, 0, 1)
			color = np.where(color <= 0.0031308, 12.92 * color, 1.055 * np.power(color, 1 / 2.4) - 0.055)
			return np.round(color * 255).astype(int)

		r, g, b = convert(r), convert(g), convert(b)
		return r, g, b

	x, y, z = lab_to_xyz(L, a, b)
	r, g, b = xyz_to_rgb(x, y, z)
	return f"#{r:02x}{g:02x}{b:02x}"

def checkviz_cmap(dist_false, dist_missing):
	## generate a 2D continous colormap for the missing and false distortion
	"""
	Generate a matplotlib color code that reproduces the above js code
	"""
	cScale = 1.3
	dist_false = 1 - dist_false
	dist_missing = 1 - dist_missing

	powScale = lambda x: x ** 1.5145
	aScale = lambda x: 30 * cScale * x
	bScale = lambda x: 20 * cScale * x

	lab = [powScale(1 - (dist_false + dist_missing) / 2) * 100, aScale(dist_false - dist_missing), bScale(dist_missing - dist_false)]
	## change the cielab color to rgb that can be used by matplotlib
	# print(lab)
	color = cielab_to_rgb_hex(*lab)
	return color


import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import Voronoi, voronoi_plot_2d
from .colormap import checkviz_cmap

from sklearn.neighbors import kneighbors_graph



def checkviz(
	scatter_data, false_distortion_list, missing_distortion_list, 
	ax=None, point_c="black", point_s=1, point_alpha=0.5, point_marker="o",
):
	vor = Voronoi(scatter_data)
	## set size
	if ax is None:
		fig, ax = plt.subplots(figsize=(10, 10))

	voronoi_plot_2d(vor, ax=ax, show_vertices=False, show_points=False,  line_width=0)

	for idx, region in enumerate(vor.regions[:-1]):
		if not -1 in region:
			polygon = [vor.vertices[i] for i in region]
			ax.fill(*zip(*polygon),  checkviz_cmap(false_distortion_list[idx], missing_distortion_list[idx]) )

	ax.scatter(scatter_data[:, 0], scatter_data[:, 1], c=point_c, zorder=2, s=point_s, alpha=point_alpha, marker=point_marker)

	ax.set_xticks([])
	ax.set_yticks([])

	



def reliability_map(
		emb, false_distortion_list, missing_distortion_list, k=7, ax=None,
		point_c="black", point_s=1, point_alpha=0.5, point_marker="o",
		linewidth=2.5, line_alpha=0.8

	):
	## construct a knn graph
	if ax is None:
		fig, ax = plt.subplots(figsize=(10, 10))

	knn_graph = kneighbors_graph(emb, k, mode="distance", include_self=False)

	## visualizae points and knn graph
	ax.scatter(emb[:, 0], emb[:, 1], c=point_c, zorder=2, s=point_s, alpha=point_alpha, marker=point_marker)
	for i in range(emb.shape[0]):
		for j in knn_graph[i].indices:
			color = checkviz_cmap((false_distortion_list[i] + false_distortion_list[j]) / 2, (missing_distortion_list[i] + missing_distortion_list[j]) / 2)
			ax.plot(
				[emb[i, 0], emb[j, 0]], [emb[i, 1], emb[j, 1]], 
				c=color, zorder=1, linewidth=linewidth, alpha=line_alpha
			)


	ax.set_xticks([])
	ax.set_yticks([])


from sklearn.metrics import silhouette_score, calinski_harabasz_score, davies_bouldin_score

def measure(emb, label, measure="silhouette"):
  """
  Compute internal validation measure of the embedding
  INPUT:
		ndarray: emb: embedded data
		str: measure: internal validation measure to compute (Optional)
			Currently supports "silhouette", "calinski_harabasz", "davies_bouldin"
  OUTPUT:
		dict: internal validation measure value
	"""
  
  if measure == "silhouette":
    score = silhouette_score(emb, label)
  elif measure == "calinski_harabasz":
    score = calinski_harabasz_score(emb, label)
  elif measure == "davies_bouldin":
    score = davies_bouldin_score(emb, label)
  
  return {
    measure: score
	}



from .utils import knn
import numpy as np

def measure(emb, label, k=20, knn_emb_info=None, return_local=False):
  """
	Compute neighborhood hit of the embedding
	INPUT:
		ndarray: emb: embedded data
		ndarray: label: label of the original data
		int: k: number of nearest neighbors to consider
		tuple: knn_info: precomputed k-nearest neighbors of the original and embedded data (Optional)
	OUTPUT:
		dict: neighborhood hit (nh)
	"""
  if knn_emb_info is None:
    emb_knn_indices = knn.knn(emb, k)
  else:
    emb_knn_indices = knn_emb_info

  points_num  = emb.shape[0]
  nh_list = []
  for i in range(points_num):
    emb_knn_index = emb_knn_indices[i]
    emb_knn_index_label = label[emb_knn_index]
    nh_list.append(np.sum((emb_knn_index_label == label[i]).astype(int)))

  nh_list = np.array(nh_list)
  nh_list = nh_list / k
  
  nh = np.mean(nh_list)
  
  if return_local:
    return ({
			"neighborhood_hit": nh
		}, {
			"local_neighborhood_hit": nh_list
		})
  else:
    return {
			"neighborhood_hit": nh
		}
  w


import numpy as np 
from .utils import knn


def measure(orig, emb, k=20, knn_ranking_info=None, return_local=False):
	"""
	Compute Mean Relative Rank Error (MRRE) of the embedding
	INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
		int: k: number of nearest neighbors to consider
		tuple: knn_ranking_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
	OUTPUT:
		dict: MRRE_false and MRRE_missing
	"""
	if knn_ranking_info is None:
		orig_knn_indices, orig_ranking = knn.knn_with_ranking(orig, k)
		emb_knn_indices,  emb_ranking  = knn.knn_with_ranking(emb, k)
	else:
		orig_knn_indices, orig_ranking, emb_knn_indices, emb_ranking = knn_ranking_info

	if return_local:
		mrre_false, local_mrre_false = mrre_computation(orig_ranking, emb_ranking, emb_knn_indices, k, return_local)
		mrre_missing, local_mrre_missing = mrre_computation(emb_ranking, orig_ranking, orig_knn_indices, k, return_local)
		return ({
			"mrre_false": mrre_false,
			"mrre_missing": mrre_missing
		}, {
			"local_mrre_false": local_mrre_false,
			"local_mrre_missing": local_mrre_missing
		})
	else:
		mrre_false = mrre_computation(orig_ranking, emb_ranking, emb_knn_indices, k, return_local)
		mrre_missing = mrre_computation(emb_ranking, orig_ranking, orig_knn_indices, k, return_local)

		return {
			"mrre_false": mrre_false,
			"mrre_missing": mrre_missing,
		}

def mrre_computation(base_ranking,target_ranking, target_knn_indices, k, return_local=False):
	"""
	Core computation of MRRE
	"""
	local_distortion_list = []
	points_num = target_knn_indices.shape[0]
	for i in range(points_num):
		base_rank_arr   = base_ranking[i][target_knn_indices[i]]
		target_rank_arr = target_ranking[i][target_knn_indices[i]]
		local_distortion_list.append(np.sum(np.abs(base_rank_arr - target_rank_arr) / target_rank_arr))
	
	c = sum([abs(points_num - 2 * i + 1) / i for i in range(1, k + 1)])
	local_distortion_list = np.array(local_distortion_list)
	local_distortion_list = 1 - local_distortion_list / c

	average_distortion = np.mean(local_distortion_list)

	if return_local:
		return average_distortion, local_distortion_list
	else:
		return average_distortion
	


	
	



import numpy as np
from numpy.linalg import svd
from .utils import pairwise_dist as pdist
from .utils import knn

def measure(orig, emb, k=20, knn_info=None):
	"""
	Compute procrustes statistics
	INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
	      int: k: number of nearest neighbors to consider
	OUTPUT:
		normalized procrustes statistics for 
	"""
	# k nearest neighbors in original space and embedded space each
	if knn_info is None:
		orig_knn_indices = knn.knn(orig, k)
		emb_knn_indices  = knn.knn(emb, k)
	else:
		orig_knn_indices, emb_knn_indices = knn_info

	g_list = []

	for i in range(orig.shape[0]):
		origin_neighbors = orig[orig_knn_indices[i]]
		embedd_neighbors = emb[emb_knn_indices[i]]

		k = origin_neighbors.shape[0]
		I = np.eye(k)
		ones = np.ones((k,k))
		H = I - (1/k) * ones

		U, _, V_T = svd(origin_neighbors.T @ H @ embedd_neighbors)
		A = U @ V_T
		A_T = A.T

		g = np.linalg.norm(H @ (origin_neighbors - embedd_neighbors @ A_T), ord='fro') ** 2
		g_normalized = g / np.linalg.norm(H @ origin_neighbors, ord='fro') ** 2
		g_list.append(g_normalized)

	return np.mean(g_list)


from .utils import pairwise_dist as pdist
import numpy as np

def measure(orig, emb, sigma=0.1, distance_matrices=None):
  """
  Compute Kullback-Leibler divergence of the embedding
  INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
		tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
  OUTPUT:
		dict: Kullback-Leibler divergence (kl)
	"""
  
  if distance_matrices is None:
    orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
    emb_distance_matrix = pdist.pairwise_distance_matrix(emb)
  else:
    orig_distance_matrix, emb_distance_matrix = distance_matrices

  density_orig = pdist.distance_matrix_to_density(orig_distance_matrix, sigma)
  density_emb = pdist.distance_matrix_to_density(emb_distance_matrix, sigma)
  
  kl = np.sum(density_orig * np.log(density_orig / density_emb)) 
  return {
    "kl_divergence": kl
	}


from scipy.stats import pearsonr
from .utils import pairwise_dist as pdist

def measure(orig, emb, distance_matrices=None):
  """
	Compute Pearson's correlation coefficient of the embedding
	INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
		tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
	OUTPUT:
		dict: Pearson's correlation coefficient (r)
  """
  
  if distance_matrices is None:
    orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
    emb_distance_matrix = pdist.pairwise_distance_matrix(emb)
  else:
    orig_distance_matrix, emb_distance_matrix = distance_matrices

  r, p = pearsonr(orig_distance_matrix.flatten(), emb_distance_matrix.flatten())
  return {
		"pearson_r": r,
	}
  



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
	long_description = fh.read()

print(setuptools.find_packages(where="src"))

setuptools.setup(
	name="zadu",
	version="0.0.6",
	author="Hyeon Jeon",
	author_email="hj@hcil.snu.ac.kr",
	description="A Python Toolkit for Evaluating the Reliability of Dimensionality Reduction Embeddings",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/hj-n/zadu",
	classifiers=[
			"Programming Language :: Python :: 3",
			"License :: OSI Approved :: MIT License",
			"Operating System :: OS Independent",
	],
	install_requires=[
		"numpy",
		"scikit-learn",
		"scipy",
		"snc",
		"matplotlib"
	],
	package_dir={"": "src"},
	packages=["zadu", "zaduvis", "zadu.measures", "zadu.measures.utils"],
	python_requires=">=3.9.0",
)


import numpy as np
from .utils import knn


def measure(orig, emb, k=20, knn_info=None, return_local=False):
  """
  Compute the local continuity meta-criteria of the embedding
  INPUT:
		ndarray: orig: original data
		ndarray: emb: embedded data
		int: k: number of nearest neighbors to consider
		tuple: knn_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
  OUTPUT:
		dict: local continuity meta-criteria
  """
  if knn_info is None:
    orig_knn_indices = knn.knn(orig, k)
    emb_knn_indices = knn.knn(emb, k)
  else:
    orig_knn_indices, emb_knn_indices = knn_info

  point_num = orig.shape[0]
  local_distortion_list = []
  
  for i in range(point_num):
    local_distortion_list.append(np.intersect1d(orig_knn_indices[i], emb_knn_indices[i]).shape[0] - ((k * k) / (point_num - 1)))

  local_distortion_list = np.array(local_distortion_list)
  local_distortion_list = local_distortion_list / k 
  
  average_distortion = np.mean(local_distortion_list)
  
  if return_local:
    return ({
      "lcmc": average_distortion
		}, {
      "local_lcmc": local_distortion_list
		})
  else:
    return {
      "lcmc": average_distortion
		}

	
      
  
  

     
